//
//	(c) Paul Campbell 2012 - released under LGPL
//	This file is part of the DSPace Robot support software.
//
//	The DSPace Robot support software is free software: you can
// 	redistribute it and/or modify it under the terms of the
//	Lesser GNU General Public License as published by the Free
//	Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	The DSPace Robot support software is distributed in the hope
//	that it will be useful, but WITHOUT ANY WARRANTY; without even
//	the implied warranty of MERCHANTABILITY or FITNESS FOR A
//	PARTICULAR PURPOSE.  See the Lesser GNU General Public License
//	for more details.
//
//	You should have received a copy of the lesser GNU General Public
//	License along with the DSPace Robot support software.  If not,
//	see <http://www.gnu.org/licenses/>.
//
#ifndef _ROBOT_H_
#define _ROBOT_H_
#include <Arduino.h>
#if ROBOT_VERSION==10
static const uint8_t analog1[] = { 0, 1, 3, 2};
static const uint8_t digital0[] = { 14, 15, 17, 16};
static const uint8_t digital1[] = { 3, 2, 18, 19};
static const uint8_t digital2[] = { 3, 2, 18, 19};  // no real port 2 on ver 0 
static const uint8_t motor_enable[] = { 4, 5};
static const uint8_t motor_i1[] = { 6, 9};
static const uint8_t motor_i2[] = { 7, 10};
#elif ROBOT_VERSION==11
static const uint8_t analog1[] = { 0, 1, 2, 3};
static const uint8_t digital0[] = { 14, 15, 16, 17};
static const uint8_t digital1[] = { 8, 2, 19, 3};
static const uint8_t digital2[] = { 18, 18, 18, 18};  // no real port 2 on ver 0 
static const uint8_t motor_enable[] = { 4, 5};
static const uint8_t motor_i1[] = { 6, 9};
static const uint8_t motor_i2[] = { 7, 10};
#elif ROBOT_VERSION==20
static const uint8_t analog1[] = { 0, 1, 2, 3, 0, 0};
static const uint8_t digital0[] = { 14, 15, 16, 17, 12, 8 };
static const uint8_t digital1[] = { 8, 2, 19, 3, 11, 13};
static const uint8_t digital2[] = { 18, 18, 18, 18, 18, 18};  // no real port 2 on ver 0 
static const uint8_t motor_enable[] = { 4, 5};
static const uint8_t motor_i1[] = { 6, 9};
static const uint8_t motor_i2[] = { 7, 10};
#else
#error "ROBOT_VERSION invalid or not defined"
#endif
typedef enum SensorPin {
    Pin0=0 ,
    Pin1=1,
    Pin2=2,
} SensorPin;

static const uint8_t servo_port[] = {5, 5, 5};
static const SensorPin servo_pin[] = {Pin0, Pin1, Pin2};

//
// sensor abstraction classes
//
//
//	The simplest class RobotSensor defines one sensor port - there are 6 ports
//
//      Ports 0-3 are connected to at least one analog pin, 4-5 digital only - analog pins can be used as digital pins too
//      If you have an application that just needs one pin you can split a port into 2
//
//      Each port has 2 unique pins and 1 shared pin (SDA - hardware i2c data)
//
//      Each port has 5 pins:
//
//        1 - PIN1 - digital
//        2 - PIN0 - analog or digital
//        3 - +5Volts
//        4 - GND
//        5 - PIN2 - digital common
//
//


class RobotSensor {  // abstraction for a sensor block 
public:
	RobotSensor(int n) { number = n; pinMode(digital1[n], INPUT); }
		
	int AnalogRead() { return analogRead(analog1[number]); }
	int DigitalRead(SensorPin num) { return digitalRead(num==Pin0?digital0[number]:num==Pin1?digital1[number]:digital2[number]); }
	void DigitalWrite(SensorPin num, int val) { digitalWrite(num==Pin0?digital0[number]:num==Pin1?digital1[number]:digital2[number], val); }
#ifdef INPUT_PULLUP
	void PinMode(SensorPin num, int val) { pinMode(num==Pin0?digital0[number]:num==Pin1?digital1[number]:digital2[number], val); }
#else
#define INPUT_PULLUP 8
	void PinMode(SensorPin num, int val) { pinMode(num==Pin0?digital0[number]:num==Pin1?digital1[number]:digital2[number], val==INPUT_PULLUP?INPUT:val); if (val==INPUT_PULLUP) DigitalWrite(num, HIGH); }
#endif
	long PulseIn(SensorPin num, int val) { return pulseIn(num==Pin0?digital0[number]:num==Pin1?digital1[number]:digital2[number], val); }
	
private:
	unsigned char	number;
};

class RobotSensor1Pin: RobotSensor {    /// abstraction for 1 pin of a sensor block
public:
	RobotSensor1Pin(int n): RobotSensor(n) { pin=Pin0;} 
	RobotSensor1Pin(int n, SensorPin p): RobotSensor(n) { pin=p;} 
	int DigitalRead() { return RobotSensor::DigitalRead(pin); }
	void DigitalWrite(int val) { RobotSensor::DigitalWrite(pin, val);}
	void PinMode(int val) { RobotSensor::PinMode(pin, val);}
	long PulseIn(int val) { RobotSensor::PulseIn(pin, val);}

private:
	SensorPin 	pin;
};

class ServoRobotSensor: public RobotSensor1Pin { // a sensor that uses any sensor pin to control a servo
public: 
	ServoRobotSensor(int n, SensorPin p=Pin0):RobotSensor1Pin(n,p) { PinMode(OUTPUT); DigitalWrite(0); Pulse(128);}
	void Pulse(int n);	// 0 = left 155 = right 128 = middle
	static void InterPulse();
	void FullPulse(int n);
};

class ServoRobot: public ServoRobotSensor { // a sensor that uses any sensor pin to control a servo
public:
	ServoRobot(int n):ServoRobotSensor(servo_port[n-1], servo_pin[n-1]) { }
};

class MicroSwitchRobotSensor: public RobotSensor1Pin { // a sensor that uses a single pin with a microswitch to ground
public:
        MicroSwitchRobotSensor(int n, SensorPin p=Pin0): RobotSensor1Pin(n,p) {PinMode(INPUT); DigitalWrite(1);} // enable pullup
        bool On() { return DigitalRead()!=1; }
};

class DIPSwitchRobotSensor: public MicroSwitchRobotSensor { // the two microswitches on the board
public:
#if ROBOT_VERSION==10 
        DIPSwitchRobotSensor(int n): MicroSwitchRobotSensor(n==1?2:1, Pin1) {}
#elif ROBOT_VERSION==11
        DIPSwitchRobotSensor(int n): MicroSwitchRobotSensor(n==1?1:0, Pin1) {}
#else
        DIPSwitchRobotSensor(int n): MicroSwitchRobotSensor(n==1?1:0, Pin1) {}
#endif
};

class LEDRobotSensor: public RobotSensor1Pin {	// an external LED with resistor to +5V
public:
	LEDRobotSensor(int n, SensorPin p=Pin0): RobotSensor1Pin(n,p) {PinMode(OUTPUT);DigitalWrite(HIGH);} 
	void On(bool n) { DigitalWrite(n?LOW:HIGH); }
	void On() { On(true); }
	void Off() { On(false); }
};

class TransistorRobotSensor: public RobotSensor1Pin {	// an external transistor driving a load
public:
	TransistorRobotSensor(int n, SensorPin p=Pin0): RobotSensor1Pin(n,p) {PinMode(OUTPUT);DigitalWrite(LOW);} 
	void On(bool n) { DigitalWrite(n?HIGH:LOW); }
	void On() { On(true); }
	void Off() { On(false); }
};


class I2CRobotSensor: public RobotSensor { // a sensor that uses the analog pin as a clock and the digital pin as data
public:
      I2CRobotSensor(int n, int w=0):RobotSensor(n){pc = (w?Pin1:Pin0); pd = (w?Pin0:Pin1);};
      void begin(int dd=5);
      void Start();
      void Stop();
      int ReadBit();
      void WriteBit(int b);
      int WriteByte(int v);
      int ReadByte();
      int WriteArray(int a, unsigned char *v, int l);
      int ReadArray(int a, unsigned char *v, int l);
      int ReadArray(int a1, int a2, unsigned char *v, int l);
      int ReadAddress(int a);
      int ReadAddress(int a1, int a2);
      int WriteAddress(int a, int v);
      int WriteAddress(int a1, int a2, int v);
      void SetDelay(int v) {d=v;}
private:
      void Data0() {DigitalWrite(pd, 0);PinMode(pd, OUTPUT); }
      void Data1() {PinMode(pd, INPUT_PULLUP); }
      void Pullup() {PinMode(pd, INPUT_PULLUP); }
      void Clock0() { DigitalWrite(pc, 0); PinMode(pc, OUTPUT); }
      void Clock1() {PinMode(pc, INPUT_PULLUP);}
      void Delay() { volatile int x=d;while (x--); }
      int d;
      SensorPin pd, pc;
};

//
//  Motor abstraction classes
//
class RobotMotor {
public:
    
	inline RobotMotor() {}
	inline void Init(int n) { number = n; pinMode(motor_enable[n], OUTPUT); digitalWrite(motor_enable[n], 0); pinMode(motor_i1[n], OUTPUT); pinMode(motor_i2[n], OUTPUT); }
	inline RobotMotor(int n) { Init(n); }
	inline void Stop() { digitalWrite(motor_enable[number], 0); }
	inline void Brake() { digitalWrite(motor_i1[number], 1);digitalWrite(motor_i2[number], 1);digitalWrite(motor_enable[number], 1); }
	inline void Forwards() { digitalWrite(motor_i1[number], 0);digitalWrite(motor_i2[number], 1);digitalWrite(motor_enable[number], 1); }
	inline void Backwards() { digitalWrite(motor_i1[number], 1);digitalWrite(motor_i2[number], 0);digitalWrite(motor_enable[number], 1); }
private:
	unsigned char number;
};

class RobotDrive {
public:
        RobotDrive() { leftMotor.Init(1); rightMotor.Init(0); }
	RobotMotor	rightMotor;
	RobotMotor	leftMotor;
	void 	  Stop() { leftMotor.Stop(); rightMotor.Stop(); }
	void 	  Forwards() { leftMotor.Forwards(); rightMotor.Forwards(); }
	void 	  Backwards() { leftMotor.Backwards(); rightMotor.Backwards(); }
	void 	  Left() { leftMotor.Backwards(); rightMotor.Forwards(); }
	void 	  Right() { leftMotor.Forwards(); rightMotor.Backwards(); }
	void 	  LeftPivot() { leftMotor.Brake(); rightMotor.Forwards(); }
	void 	  RightPivot() { leftMotor.Forwards(); rightMotor.Brake(); }
};

#endif
